/*
** hash.h - hash table header
**
** Keith Zaback (krz@wcug.wwu.edu)
*/


#ifndef INC_HASH_H
#define INC_HASH_H

#define HASH_MAGIC		0x47405347	/* magic number "HASH" */

typedef struct stHASHENTRY {
	void			*data;		/* their data */
	struct stHASHENTRY	*next;		/* our link */
} HASHENTRY;

typedef struct stHASHTABLE {
	int		magic;			/* stuff */
	int		total_buckets;		/* # of buckets */
	int		entries;		/* number of entries */
	int		entry_size;		/* size of entries */
	int		(*compare)(void *e1, void *e2);	/* compare function */
	int		(*hash)(void *e1);	/* hash function */
	HASHENTRY	**table;		/* the table */
} HASHTABLE;

typedef struct stHASHTRAV {
	HASHTABLE		*htab;		/* hashtable we are using */
	int			bucket;		/* bucket we are in */
	HASHENTRY		*entry;		/* entry we are on */
} HASHTRAV;

/* protos */
int		create_hash_table(HASHTABLE *htab, int total_entries, int entry_size, int (*hash)(void *), int (*compare)(void *, void *));
void		free_hash_table(HASHTABLE *htab);
void		*hash_get(HASHTABLE *htab, void *item);
void		*hash_insert(HASHTABLE *htab, void *item);
void		hash_enum_entries(HASHTABLE *htab, void (*enumfunc)(int, void *));
void		hash_empty_table(HASHTABLE *htab);
int		hash_remove(HASHTABLE *htab, void *item);
int		hash_remove_multi(HASHTABLE *htab, int (*testfunc)(void *));
void		*hash_start_trav(HASHTABLE *htab, HASHTRAV *trav);
void		*hash_next_trav(HASHTRAV *trav);
#endif

