/*
** ether_linux.c - the ethernet part of IPX fury (LINUX style)
**
** Keith Zaback (krz@wcug.wwu.edu) Mar 98
*/

/* includes */
#include "config.h"
#ifdef	FURY_ETHERNET
#ifdef	FURY_LINUX

#include "ipxfury.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>

/* defines */
#define LOG_FRAMES	0


/* externs */
extern GLOBALS	ifg;


/*
**	ETHER_FLUSH - remove any accum'd packets since we've opened
*/
int ether_flush(INTERFACE *edev)
{
	int		i, nfds;
	fd_set		readfds;
	char		buffer[FURY_ETHER_FRAME_MAX];
	struct timeval	tv;
	
	/* init the counter */
	i = 0;
	while( i < ETHER_FLUSH_MAX ) {
		/* clean fdset */
		FD_ZERO(&readfds);
		
		/* set this device */
		FD_SET(edev->fd, &readfds);
		
		/* timeout */
		tv.tv_sec = 0;
		tv.tv_usec = 250000;	/* 1/4 second */
		
		/* now select */
		nfds = select(FD_SETSIZE, &readfds, NULL, NULL, &tv);
		if( nfds == 0 ) {
			/* timeout */
			break;
		} else if( nfds > 0 ) {
			/* read in the packet */
			read(edev->fd, buffer, FURY_ETHER_FRAME_MAX);
		} else {
			errormsg(0, "ether_flush", "select() error");
			break;
		}
		/* inc counter */
		i++;
	}
#if 0 	
	/* info */
	errormsg(0, "ether_flush", "removed %d packets", i);
#endif	
	/* no real errors */
	return(1);
}





/*
**	ETHER_READ - 	reads a packet from the ethernet device and stores
**		     	it in the buffer.  Since we only get 1 frame per
**			read from this socket, the buffer only has to 
**			have enough space to hold a frame (1514).  Also,
**			the buffer should be clear everytime we get here.
*/
int ether_read_old(INTERFACE *edev)
{
	int			bytes_read;
	unsigned char		temp[FURY_ETHER_FRAME_MAX];
	IPXHEADER		*ipxheader;
	
	/* read one ethernet frame into temp buffer */
	bytes_read = read(edev->fd, temp, FURY_ETHER_FRAME_MAX);

	/* did we get anything? */
	if( bytes_read > 0 ) {
		/* set the header */
		ipxheader = ether_get_ipx(edev->data.e.e_frametype, temp);
		if( NULL == ipxheader ) {
			return(0);
		}
		
		/* log it */
#if LOG_FRAMES == 1
		fprintf(ifg.logfile, "\n>>>>  Rcvd IPX Frame: [%s]\n", current_time());
		print_ipx_frame((BYTE *)ipxheader, 0);
		fhexdump(ifg.logfile, (((BYTE *)ipxheader)+30), (ntohs(ipxheader->length) - 30));
#endif
		
		/* adjust the stats */
		edev->stats.recv_bytes_recent 	+= bytes_read;
		edev->stats.bytes_rcvd		+= bytes_read;

		/* check the port */
		if( check_port(ifg.watched_ports, ntohs(ipxheader->ipx_dest.port)) ) {
			/* add it to the buffer */
			if( ! buffer_insert(&edev->recvbuf, (BYTE *)ipxheader, ntohs(ipxheader->length)) ) {
				errormsg(0, "ether_read", "could not insert %d byte(s) into buffer", bytes_read);
			} else {
				return(1);
			}
		}
	} else {
		errormsg(0, "ether_read", "failed to read from device");
	}
	
	/* failure */
	return(0);
}

/*
**	ETHER_READ - lots of hardcoded values in here.. need to FIXTHIS
*/
int ether_read_old2(INTERFACE *edev)
{
	int			bytes_read;
	int			frames_read, frames_added;
	unsigned char		temp[FURY_ETHER_FRAME_MAX];
	fd_set			readfds;
	struct timeval		tv;
	IPXHEADER		*ipxheader;
	int			nfds, ipx_length;  
	
	/* init */
	frames_read = 0;
	frames_added = 0;
	
	/* the loop */
	do {	
		/* create fd set */
		FD_ZERO(&readfds);
		FD_SET(edev->fd, &readfds);
		
		/* set the timeout */
		tv.tv_sec = 0;
		tv.tv_usec = 250;	/* ~1000/1000000 at max */
		
		/* do the select */
		nfds = select(FD_SETSIZE, &readfds, NULL, NULL, &tv);
		if( nfds > 0 ) {
			/* no need to check since we know it is us */
			bytes_read = read(edev->fd, temp, FURY_ETHER_FRAME_MAX);
			if( bytes_read > 14 ) {
				/* adjust the stats */
				edev->stats.recv_bytes_recent += bytes_read;
					
				/* set a ptr to work with */
				ipxheader = ether_get_ipx(edev->data.e.e_frametype, temp);
				if( NULL == ipxheader ) {
					return(0);
				}
				
				/* make sure it is good, add it to the buffer */
				ipx_length = ntohs(ipxheader->length);
				if( (ipx_length < 30) || (ipx_length > 1500) ) {
					errormsg(0, "ether_read", "bad ipx frame size (%d bytes)", ipx_length);
					return(0);
				} else {
					/* should we actually deal with this port? */
					if( check_port(ifg.watched_ports, htons(ipxheader->ipx_dest.port)) ) {
						/* add it to the buffer */
						if( ! buffer_insert(&edev->recvbuf, (BYTE *)ipxheader, ipx_length) ) {
#if 0						
							errormsg(0, "ether_read", "could not add ipx frame to recvbuf (%d bytes)", ipx_length);
#endif							
							return(0);
						}
						/* inc the frame counter */
						frames_added++;					
					} else {
#if 0					
						errormsg(0, "ether_read", "blocked port! src:%d -> dest:%d", ntohs(ipxheader->ipx_src.port), ntohs(ipxheader->ipx_dest.port));
#endif					
					}
					
					/* we read a frame */
					frames_read++;
				}	
				
			}
		} /* nfds > 0 */
	
	} while( (nfds > 0) && (frames_read < ETHER_FRAMES_READ_MAX) );
	
	/* done */
	return(frames_added);

}

int ether_read(INTERFACE *edev)
{
	int			bytes_read;
	int			frames_read, frames_added;
	unsigned char		temp[FURY_ETHER_FRAME_MAX];
	IPXHEADER		*ipxheader;
	int			ipx_length, bytes_avail;  
	
	/* init */
	frames_read = 0;
	frames_added = 0;
	
	/* the loop */
	do {	
		/* no need to check since we know it is us */
		bytes_read = read(edev->fd, temp, FURY_ETHER_FRAME_MAX);
		if( bytes_read > 14 ) {
			/* adjust the stats */
			edev->stats.recv_bytes_recent += bytes_read;
				
			/* set a ptr to work with */
			ipxheader = ether_get_ipx(edev->data.e.e_frametype, temp);
			if( NULL == ipxheader ) {
				return(0);
			}
			
			/* make sure it is good, add it to the buffer */
			ipx_length = ntohs(ipxheader->length);
			if( (ipx_length < 30) || (ipx_length > 1500) ) {
				errormsg(0, "ether_read", "bad ipx frame size (%d bytes)", ipx_length);
				return(0);
			} else {
				/* should we actually deal with this port? */
				if( check_port(ifg.watched_ports, htons(ipxheader->ipx_dest.port)) ) {
					/* add it to the buffer */
					if( ! buffer_insert(&edev->recvbuf, (BYTE *)ipxheader, ipx_length) ) {
#if 0						
						errormsg(0, "ether_read", "could not add ipx frame to recvbuf (%d bytes)", ipx_length);
#endif							
						return(0);
					}
					/* inc the frame counter */
					frames_added++;					
				} else {
#if 0				
					errormsg(0, "ether_read", "blocked port! src:%d -> dest:%d", ntohs(ipxheader->ipx_src.port), ntohs(ipxheader->ipx_dest.port));
#endif				
				}
			}	
		}
		
		/* we did a read */
		frames_read++;
		
		/* is there more to read? */
		if( ioctl(edev->fd, FIONREAD, &bytes_avail) < 0 ) {
			bytes_avail = 0;
		} 
	
	} while( (bytes_avail > 0) && (frames_read < ETHER_FRAMES_READ_MAX) );
	
	/* done */
	return(frames_added);
}



/*
**	ETHER_OPEN() - this opens the sets up the ethernet interface
*/
int ether_open(INTERFACE *edev)
{	
	struct ifreq		ifr;
	int			proto;
	
	/* set proto based on frame type */
	switch( edev->data.e.e_frametype ) {
		case FURY_FRAMETYPE_II:
			proto = ETH_P_IPX;
			break;
					
		case FURY_FRAMETYPE_802_3:
			proto = ETH_P_802_3;
			break;
			
		case FURY_FRAMETYPE_802_2:
			proto = ETH_P_802_2;
			break;
			
		default:
			errormsg(0, "ether_open", "unknown frametype (%d)", edev->data.e.e_frametype);
			return(0);
	}
			
	/* get a file descriptor to talk with the ethernet card */
	edev->fd =  socket(AF_INET, SOCK_PACKET, htons(proto));
	if( edev->fd < 0 ) {
		errormsg(0, "ether_open", "socket() failed!");
		return(0);
	}
	
	/* now bind it */
	memset(&edev->data.e.e_sa, 0, sizeof(edev->data.e.e_sa));
	edev->data.e.e_sa.sa_family = AF_UNIX;
	strcpy(edev->data.e.e_sa.sa_data, edev->name);
	if( bind(edev->fd, &edev->data.e.e_sa, sizeof(edev->data.e.e_sa)) < 0 ) {
		errormsg(0, "ether_open", "bind(%s) failed!", edev->name);
		close(edev->fd);
		return(0);
	}
	
	/* get the device flags */
	strcpy(ifr.ifr_name, edev->name);
	if( ioctl(edev->fd, SIOCGIFFLAGS, &ifr) < 0 ) {
		errormsg(0, "ether_open", "ioctl(%s) SIOCGIFFLAGS failed!", edev->name);
		close(edev->fd);
		return(0);
	}
	
	/* set the proper mode */
	ifr.ifr_flags |= IFF_PROMISC;
	
	/* now set the flags */
	if( ioctl(edev->fd, SIOCSIFFLAGS, &ifr) < 0 ) {
		errormsg(0, "ether_open", "ioctl(%s) SIOCSIFFLAGS failed!", edev->name);
		close(edev->fd);
		return(0);
	}
	
	/* create the buffers */
	if( ! create_buffer(&edev->recvbuf, FURY_ETHER_RECV_BUF_SIZE) ) {
		close(edev->fd);
		return(0);
	}
	
	if( ! create_buffer(&edev->sendbuf, FURY_ETHER_TRANS_BUF_SIZE) ) {
		close(edev->fd);
		return(0);
	}
	
	/* we seem to have done some good */
	edev->flags |= FURY_IF_VALID;		/* this interface is valid */
	edev->state = FURY_STATE_INITIAL;	/* just getting started */
	
	/* set recent stats starting time */
	gettimeofday(&edev->stats.recent_stats_start, NULL);
	
	return(1);				/* return success! */
	
}


/*
**	ETHER_CLOSE()
*/
int ether_close(INTERFACE *edev)
{
	struct ifreq		ifr;
	
	/* is this interface valid? */
	if( (edev->flags & FURY_IF_VALID) != FURY_IF_VALID ) {
		errormsg(0, "ether_close", "called with non-valid interface.");
		return(0);
	}
	
	/* free the buffers */
	free_buffer(&edev->recvbuf);
	free_buffer(&edev->sendbuf);

	/* get the device flags */
	strcpy(ifr.ifr_name, edev->name);
	if( ioctl(edev->fd, SIOCGIFFLAGS, &ifr) < 0 ) {
		errormsg(0, "ether_open", "ioctl(%s) SIOCGIFFLAGS failed!", edev->name);
		close(edev->fd);
		return(0);
	}
	
	/* remove promisc mode */
	ifr.ifr_flags &= ~IFF_PROMISC;
	
	/* now set the flags */
	if( ioctl(edev->fd, SIOCSIFFLAGS, &ifr) < 0 ) {
		errormsg(0, "ether_close", "ioctl(%s) SIOCSIFFLAGS failed!", edev->name);
		return(0);
	}

	/* now actually close the file descriptor */
	close(edev->fd);
	
	/* this is no longer valid */
	edev->fd = -1;
	
	/* set the flags */	
	edev->flags &= ~FURY_IF_VALID;
	
	/* did fine */
	return(1);
	
}
	
#endif /* FURY_LINUX */

#endif /* FURY_ETHERNET */
